#
# This file is a portion of the Red Hat Update Agent
# Copyright (c) 1999 Red Hat, Inc.
#
# $Id: client.py,v 1.57 1999/11/03 20:17:22 pbrown Exp $

import sys, os, zlib
import string, urllib, httplib, smtplib, socket
import mimetools, multifile
import re
from fnmatch import fnmatch
from rpmdb import *
from threading import Thread
from gtk import *

import config

# constants
NEW = 1
DEPEND = 2
INSTALLED = 3
ERROR = 4
DOWNLOAD = 5

# HELPER FUNCTIONS
def getVersion() :
    db = RPMDatabase(0, "/")
    pkgLabels = db.packageLabels()
    for pkg in pkgLabels:
	if re.match("redhat-release", pkg):
	    hdr = db.getHeader(pkg)
	    db.close()
	    return hdr['version']

    db.close()
    return "Unknown"

def getArch() :
    s = os.uname()[4]
    s = re.sub("i.86", "i386", s)
    return s

# compare two RPM packages
def comparePackages(pkgLabel1, pkgLabel2):
    (name1, epoch1, version1, release1) = string.split(pkgLabel1)
    (name2, epoch2, version2, release2) = string.split(pkgLabel2)
    # this is a hack
    if epoch1 == "None":
	epoch1 = None
    if epoch2 == "None":
	    epoch2 = None
    return rpm.labelCompare((epoch1, version1, release1),
			    (epoch2, version2, release2))


class Client :
    def __init__(self, auth, gui=None):
	self.cfg = config.Up2dateConfig()
	self.cfg.load()

	self.bounce = 0
	if auth == "":
	    auth = self.cfg.readEntry("regKey")
	    if auth == "":
		auth = "null"
	username = self.cfg.readEntry("regName")

	self.gui = gui
	    
	self.auth = urllib.quote("%s" % auth) # make it good for sending
	self.username = urllib.quote(username)

	self.pkgList = []
	self.pkgSizes = {}
	self.state = ""
	
	self.arch = getArch()
	self.reply = ""
	self.mailmsg = ""
	
    # def sends any pending message and exits
    def finishUp(self, subject = "Successful Red Hat Update Agent run"):
	if not self.gui:
	    self.reply = "Subject: " + subject + "\n\n" + self.reply

	    server = smtplib.SMTP("localhost")
	    server.sendmail('up2date@localhost', 
			    self.cfg.readEntry("adminAddress"), self.reply)
	    server.quit()
	    sys.exit()
	else:
	    self.mailmsg = self.mailmsg + "\nSome sort of error has occurred, aborting.\n"
	    self.mailSummary()
	    self.gui.dieThreads = 1
	    return

    def mailSummary(self):
	self.mailmsg = "Subject: Results from Red Hat Update Agent\n\n" + self.mailmsg
	server = smtplib.SMTP("localhost")
	server.sendmail('root@localhost',
			self.cfg.readEntry("adminAddress"), self.mailmsg)
	server.quit()
	
    # simple text message replies
    def addReply(self, msg = "", nocr = 0, debug = 0):
	if debug and not self.cfg.readEntry("debug"):
	    return

	if not self.gui:
	    self.reply = self.reply + msg
	    if not nocr:
		self.reply = self.reply + "\n"
	else:
	    threads_enter()
	    if not nocr:
		print msg
		#self.gui.addText("%s" % msg + "\n")
	    else:
		print msg,
		#self.gui.addText("%s" % msg)
	    threads_leave()

    def setStatus(self, msg):
	if not self.gui:
	    print "STATUS: %s" % msg
	    return 0
	else:
	    if self.cfg.readEntry("debug"):
		print "STATUS: %s" % msg
	    threads_enter()
	    cid = self.gui.statusBar.get_context_id(msg)
	    self.gui.statusBar.push(cid, msg)
	    threads_leave()
	    return cid

    def setProgress(self, amount, total):
	threads_enter()
	self.gui.progressBar.set_activity_mode(FALSE)
	threads_leave()
	if not self.gui:
	    return
	else:
	    if total != 0:
		i = (amount + 1.0) / total
		if i > 1:
		    i = 1.0
		elif i < 0:
		    i = 0
	    else:
		i = 0
	    threads_enter()
	    self.gui.progressBar.update(i)
	    threads_leave()

    def setActivity(self):
	threads_enter()
	self.gui.progressBar.set_activity_mode(TRUE)
	self.bounce = not self.bounce
	self.gui.progressBar.update(self.bounce)
	threads_leave()
    
    def unsetStatus(self, cid):
	if not self.gui:
	    return
	else:
	    threads_enter()
	    self.gui.statusBar.pop(cid)
	    threads_leave()
    
    def addWarning(self, msg):
	self.addReply("WARNING: %s" % msg)

    # used to load a list of packages from disk
    def loadPkgList(self):
	None

    # used to save a list of packages out to disk
    def savePkgList(self):
	None

    def postAction(self, dataStr) :
	if self.gui and self.gui.dieThreads:
	    return (0, 0)

	try:
	    h = httplib.HTTP(self.cfg.readEntry("pkgServer"))
	except socket.error:
	    if self.gui:
		self.gui.error("Error connecting to package server.")
	    else:
		self.addReply('Error connecting to package server.')
	    self.finishUp("Failed Red Hat Update Agent run")
	    return (0, 0)
	    
	h.putrequest('POST', self.cfg.readEntry("pkgProg"))
	    
	h.putheader('Accept', '*/*')
	h.putheader('Content-length', "%s" % len(dataStr))
	h.putheader('Content-type', 'application/x-www-form-urlencoded')
	h.putheader('User-Agent', 'up2date/1.0')
	h.endheaders()
	
	h.send(dataStr)
	errcode, errmsg, headers = h.getreply()
	if errcode != 200:
	    if self.gui:
		self.gui.error("Error %s (%s) while receiving response from package server." % (errcode, errmsg))

	    else:
		self.addReply("Error %s (%s) while receiving response from package server." % (errcode, errmsg))
	    self.finishUp("Failed Red Hat Update Agent run")
	    return (0, 0)
	 
	boundary = headers.getparam("boundary")
	if boundary == None: # if no boundary, an error occurred
	    if self.gui:
		self.gui.error(h.getfile().read())
	    else:
		self.addReply(h.getfile().read())
	    self.finishUp("Failed Red Hat Update Agent run")
	    return (0, 0)

	else:
	    return (h.getfile(), boundary)


    def processState(self, f, boundary):
	msg = multifile.MultiFile(f, 0) # we can't rewind
	msg.push(boundary)
	while 1:
	    if self.gui and self.gui.dieThreads:
		return 1

	    mmsg = mimetools.Message(msg)
	    name = mmsg.getparam("name")
	    if name == "STATE":
		self.state = msg.read()
		msg.next()
		return 0

	    msg.next()

	return 0


    def logout(self):
	dataList = []
	dataList.append("command=" + 
			urllib.quote_plus("done"))
	dataList.append("username=" + urllib.quote_plus(self.username))
	dataList.append("auth=" + urllib.quote_plus(self.auth))
	dataList.append("state=" + urllib.quote_plus(self.state))
	dataStr = string.join(dataList,"&")
	#print "dataStr is", dataStr

	try:
	    h = httplib.HTTP(self.cfg.readEntry("pkgServer"))
	except socket.error:
	    self.addReply('Error connecting to package server.')
	    self.addReply("Failed Red Hat Update Agent run")
	    
	h.putrequest('POST', self.cfg.readEntry("pkgProg"))
	    
	h.putheader('Accept', '*/*')
	h.putheader('Content-length', "%s" % len(dataStr))
	h.putheader('Content-type', 'application/x-www-form-urlencoded')
	h.putheader('User-Agent', 'up2date/1.0')
	h.endheaders()
	
	h.send(dataStr)
	errcode, errmsg, headers = h.getreply()
	if errcode != 200:
	    self.addReply("Error %s (%s) while receiving response from package server." % (errcode, errmsg))
	    self.addReply("Failed Red Hat Update Agent run")
	 
	boundary = headers.getparam("boundary")
	if boundary == None: # if no boundary, an error occurred
	    self.addReply(h.getfile().read())
	    self.addReply("Failed Red Hat Update Agent run")

	
    def getPackageList(self) :
	cid = self.setStatus(" Retrieving package list...")
	self.pkgList = []
	# construct data to send
	
	dataList = []

	try:
	    version = self.cfg.readEntry("versionOverride")
	except:
	    version = getVersion()
	if version == "":
	    version = getVersion()

	dataList.append("command=" + 
			urllib.quote_plus("listall %s %s" % 
					  (version, self.arch)))
	dataList.append("username=" + urllib.quote_plus(self.username))
	dataList.append("auth=" + urllib.quote_plus(self.auth))
	dataList.append("state=" + urllib.quote_plus(self.state))
	dataStr = string.join(dataList,"&")
	#print "dataStr is", dataStr

	(fp, boundary) = self.postAction(dataStr)

	if fp == 0:
	    return

	if self.processState(fp, boundary):
	    return
	
	msg = multifile.MultiFile(fp, 0)
	msg.push(boundary)
	mmsg = mimetools.Message(msg)
	lines = msg.readlines()
	lineCount = len(lines)
	for i in range(lineCount):
	    if self.gui and self.gui.dieThreads:
		return

	    self.setProgress(i, lineCount)
	    try:
		(dist, name, epoch, version, release) = string.split(lines[i])
	    except:
		continue
	    self.pkgList.append("%s %s %s %s" % 
				(name, epoch, version, release))

	fp.close()
	self.unsetStatus(cid)


    # this method will remove those packages that are already
    # installed.  It will also remove packages from the list
    # if they are older than a version that is currently installed.
    def removeInstalledPackages(self, pkgList, retrieveNew) :
	cid = self.setStatus(" Determining what to do with packages...")
	self.setProgress(0, 0)

	self.db = RPMDatabase(0, "/")
	labels = self.db.packageLabels()
	removedList = []

	pkgCount = len(pkgList)
	for i in range(pkgCount):
	    if self.gui and self.gui.dieThreads:
		return

	    self.setProgress(i, pkgCount)
	    for label in labels:
		pkgName = string.split(pkgList[i])[0]
		labelName = string.split(label)[0]
		if pkgName == labelName:
		    ret = comparePackages(pkgList[i], label)
		    if ret == 0:
			# already installed
			removedList.append(pkgList[i])
		    elif ret < 0:
			self.addReply("removing, a newer version is installed.", 0, 1)
			removedList.append(pkgList[i])
		    else:
		        # we'll retrieve it
			pass
		    break
	    else:
		if retrieveNew:
		    # mark this new package for retrieval implicitly
		    pass
		else:
		    # not installed, we don't want it
		    removedList.append(pkgList[i])
		
       	self.db.close()

	for rmPkg in removedList:
	    pkgList.remove(rmPkg)
	
	self.unsetStatus(cid)

	
    def removeSkipPackages(self, pkgList):
	cid = self.setStatus(" Removing packages marked to be skipped...")
	removedList = []
	for pkg in pkgList:
	    if self.gui and self.gui.dieThreads:
		return

	    (name, epoch, version, release) = string.split(pkg)
	    for pat in self.cfg.readEntry("pkgSkipList"):
		if fnmatch(name, pat):
		    self.addReply("skipping package %s cause of package rule match." % pkg, 0, 1)
		    removedList.append(pkg)
		    break

	for pkg in removedList:
	    pkgList.remove(pkg)
	    
	self.unsetStatus(cid)
	return removedList

    def removeSkipFiles(self, pkgList):
	cid = self.setStatus(" Removing packages with files marked to be skipped...")
	removedList = []
	self.setProgress(0, 0)

	self.db = RPMDatabase(0, "/")
	labels = self.db.packageLabels()

	pkgCount = len(pkgList)
	for p_i in range(pkgCount):
	    if self.gui and self.gui.dieThreads:
		return

	    pkgName = string.split(pkgList[p_i])[0]
	    fd = open("%s/%s.hdr" % 
		      (self.cfg.readEntry("storageDir"), 
		       re.sub(" ", "-", pkgList[p_i], "r")))
	    hdr = rpm.headerLoad(fd.read())
	    fd.close()
	    fflag = 0
	    fileNames = hdr['filenames']
	    fileFlags = hdr['fileflags']
	    for f_i in range(len(fileNames)):
		self.setActivity()
		if self.gui and self.gui.dieThreads:
		    return

		for pat in self.cfg.readEntry("fileSkipList"):
		    if fnmatch(fileNames[f_i], pat):
			self.addReply("skipping package %s because of file rule match." % pkgList[p_i], 0, 1)
			removedList.append(pkgList[p_i])
			fflag = 1
			break

		# Deal with config files
		if self.cfg.readEntry("noReplaceConfig"):
		    if fileFlags[f_i] & rpm.RPMFILE_CONFIG:
			# check to see if this file is different from the one
			# on disk.
			MD5Res = ""
			for label in labels:
			    labelName = string.split(label)[0]
			    if pkgName == labelName:
				installedHdr = self.db.getHeader(label)
				installedFileNames = installedHdr['filenames']
				installedFileMD5s = installedHdr['filemd5s']
				installedHdr.unload()
				for j in range(len(installedFileNames)):
				    if fnmatch(installedFileNames[j], 
					       fileNames[f_i]):
					self.addReply("matched config file %s, package %s" % (installedFileNames[j], pkgName), 0, 1)
					MD5Res = installedFileMD5s[j]
			    
	
			if MD5Res != "":
			    import md5
			    hash = md5.new()
			    try:
				f = open(fileNames[f_i], 'r')
			    except:
				self.addReply("skipping package %s because we cannot read the file from disk, and it contains config files." % pkgList[p_i], 0, 1)
				removedList.append(pkgList[p_i])
				fflag = 1
				break

			    ln = f.readline()
			    while ln:
				hash.update(ln)
				ln = f.readline()
			    f.close()

			    h = string.hexdigits
			    r = ''
			    for c in hash.digest():
				i1 = ord(c)
				r = r + h[(i1 >> 4) & 0xF] + h[i1 & 0xF]
				
			    if MD5Res != r:
                                self.addReply("skipping package %s because config file %s is locally modified." % (pkgList[p_i], fileNames[f_i]), 0, 1)
                                removedList.append(pkgList[p_i])
				fflag = 1
                # end of config file processing		
		if fflag:
		    break

	self.db.close()

	for pkg in removedList:
	    name = string.split(pkg)[0]
	    filename = ("%s/%s.hdr" % 
			(self.cfg.readEntry("storageDir"), re.sub(" ", "-", pkg)))
	    os.remove(filename)
	    pkgList.remove(pkg)
	    if self.gui:
		self.gui.deletePkgArea(name)

	self.unsetStatus(cid)
	return removedList

    def retrieveHeaders(self, pkgList):
	badPkgList = []
	cid = self.setStatus(" Retrieving package headers for dependency resolution...")
	self.setProgress(0, 0)

	pkgCount = len(pkgList)
	for i in range(pkgCount):
	    if self.gui and self.gui.dieThreads:
		return

	    name = string.split(pkgList[i])[0]
	    hcid = self.setStatus(" Retrieving header %s..." % name)

	    # may already be on disk
	    if os.access("%s/%s.hdr" % 
			 (self.cfg.readEntry("storageDir"),
			  re.sub(" ", "-", pkgList[i])), 
			 os.R_OK):
		self.setProgress(i, pkgCount)
		self.unsetStatus(hcid)
		continue

	    # construct data to send
	    dataList = []
	    dataList.append("command=" + urllib.quote_plus("header %s" % pkgList[i]))
	    dataList.append("username=" + urllib.quote_plus(self.username))
	    dataList.append("auth=" + urllib.quote_plus(self.auth))
	    dataList.append("state=" + urllib.quote_plus(self.state))
	    dataStr = string.join(dataList, "&")
	    #self.addReply("dataStr is %s" % dataStr, 0, 1)

	    (fp, boundary) = self.postAction(dataStr)
	    if fp == 0:
		return

	    if self.processState(fp, boundary):
		return
	    
	    
	    msg = multifile.MultiFile(fp, 0)
	    msg.push(boundary)
	    mmsg = mimetools.Message(msg)
	    #print ("Content length of %s: %s" 
	    #   % (pkgList[i], mmsg.get("Content-Length")))
	    try:
		self.pkgSizes[pkgList[i]] = string.atoi(mmsg.get("X-Package-Size"))
	    except:
		self.pkgSizes[pkgList[i]] = 0

	    if mmsg.gettype() == "application/octet-stream":
		ofp = os.open("%s/%s.hdr" % 
			      (self.cfg.readEntry("storageDir"), 
			       re.sub(" ", "-", pkgList[i])), 
			      os.O_WRONLY|os.O_CREAT, 0600)
		hdrData = mmsg.fp.read()

		if mmsg.get("Content-Encoding") == "x-gzip":
		    hdrData = zlib.decompress(hdrData)

		os.write(ofp, hdrData)
		os.close(ofp)
		#self.addReply("retrieved.", 0, 1)
	    else:
		#self.addReply("FAILED.", 0, 1)
		badPkgList.append(pkgList[i])
		name = string.split(pkgList[i])[0]
		if self.gui:
		    self.gui.updatePkgArea(name, ERROR)

	    self.setProgress(i, pkgCount)
	    fp.close()	    
	    self.unsetStatus(hcid)

	# remove those packages for which we couldn't retrieve a header
	for pkg in badPkgList:
	    pkgList.remove(pkg)
	    if self.gui:
		name = string.split(pkg)[0]
		self.gui.deletePkgArea(name)

	self.unsetStatus(cid)


    def checkFreeDiskSpace(self):
	cid = self.setStatus(" Checking free disk space...")
	import statvfs
	totalSize = 0
	for pkg in self.pkgList:
	    if self.gui and self.gui.dieThreads:
		return

	    try:
		totalSize = totalSize + self.pkgSizes[pkg]
	    except:
		pass

	totalSize = totalSize / 1024

	self.addReply("total size of packages: %s kbytes" % totalSize, 0, 1)
	dfInfo = os.statvfs(self.cfg.readEntry("storageDir"))
	self.addReply("free space available: %s kbytes" % dfInfo[statvfs.F_BAVAIL], 0, 1)
	if dfInfo[statvfs.F_BAVAIL] < totalSize:
	    if not self.gui:
		self.addReply("Not enough free disk space to proceed.")
	    else:
		self.gui.error("Not enough free disk space to proceed.")

	    self.finishUp("Not enough free disk space")
	    return

	self.unsetStatus(cid)


    # necessary callback for RPM installation transactions
    def callBack(self, what, amount, total, key, data):
	if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
	    d = os.open(key, os.O_RDONLY)
	    return d
	elif what == rpm.RPMCALLBACK_INST_START:
	    self.setStatus(" Installing %s..." % key)
	elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
	    name = string.split(key, "/")[-1:][0]
	    name = string.split(name, "-")[:-2]
	    name = string.join(name, "-")
	    self.gui.updatePkgArea(name, INSTALLED)
	elif what == rpm.RPMCALLBACK_INST_PROGRESS:
	    i = int((amount * 1.0) / total)
	    self.setProgress(amount, total)

    def callBack1(self,what,amount,total,key,data):
	if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
	    d = os.open(key, os.O_RDONLY)
	    return d
	elif what == rpm.RPMCALLBACK_TRANS_PROGRESS:
	    i = int((amount * 1.0) / total)
	    self.setActivity()

    def dryRun(self):
	cid = self.setStatus(" Building transaction...")
	
	for i in range(1, self.cfg.readEntry("depTries")+1): # dependency loop
	    pkgCount = len(self.pkgList)

	    self.db = RPMDatabase(0, "/")
	    ts = rpm.TransactionSet("/", self.db.db)

	    for j in range(pkgCount):
		self.setProgress(j, pkgCount)
		if self.gui and self.gui.dieThreads:
		    return

		fd = open("%s/%s.hdr" % 
			  (self.cfg.readEntry("storageDir"), 
			   re.sub(" ", "-", self.pkgList[j], "r")))
		hdr = rpm.headerLoad(fd.read())
		fd.close()
		ts.add(hdr, self.pkgList[j], "u")

	    ts.order()

	    dcid = self.setStatus(" Dependency check, pass %s..." % i)
	    deps = ts.depcheck()
	    self.unsetStatus(dcid)
	    if deps:
		for depInfo in deps:
		    self.resolveDep(depInfo)
	    else:
		self.addReply("All dependencies resolved.", 0, 1)
		break
	self.unsetStatus(cid)

	if i == self.cfg.readEntry("depTries"):
	    if not self.gui:
		self.addReply("Could not resolve dependencies.  Either the packages")
		self.addReply("required are unavailable, or your 'depTries' configuration")
		self.addReply("option needs to be increased.")
	    else:
		self.gui.error("Could not resolve dependencies.  Either the packages\n"
			   "required are unavailable, or your 'depTries' configuration\n"
			   "option needs to be increased.")

	    self.finishUp("Failed Red Hat Update Agent run")
	    return

	# remove hdrs
	for pkg in self.pkgList:
	    filename = ("%s/%s.hdr" % 
			(self.cfg.readEntry("storageDir"), 
			 re.sub(" ", "-", pkg)))
	    os.remove(filename)

	if self.gui and self.gui.dieThreads:
	    return

	cid = self.setStatus(" Dry transaction run...")
	self.setProgress(0, 0)
	rc = ts.run(rpm.RPMTRANS_FLAG_BUILD_PROBS, 0, self.callBack1, "arg")
	self.db.close()
	self.setProgress(0, 0)

	if (len(rc) != 0):
	    if not self.gui:
		for msg in rc:
		    self.addReply(msg[0])
	    else:
		mmsg = ""
		for msg in rc:
		    mmsg = mmsg + "%s\n" % msg[0]
		self.gui.error(mmsg)

	    self.finishUp("Bad results from dry transaction run")
	    return
		
	self.unsetStatus(cid)

    def resolveDep(self, depInfo):
 	((name, version, release),
 	 (needsName, needsVersion),
 	 flags,suggested,sense) = depInfo

	def getName(label):
	    s = string.split(label)
	    return s[0]

	#dblabels = self.db.packageLabels()
	#dbnames = map(getName, dblabels)

	#if flags:
	    #dblabels = db.packageLabels()
	    #fakeLabel = "%s None %s 0" % (needsName, needsVersion)
	    #for label in dblabels:
		#if string.split(label)[0] == needsName:
		    # the package we need is already on the system,
		    # see if we really need to upgrade it.
		    #if flags & rpm.RPMSENSE_GREATER:
			#if flags & rpm.RPMSENSE_EQUAL:
			    #if comparePackages(label, fakeLabel) < 0:
				# we already have a version which satisfies
				# the dependency, we can safely skip adding
				# this package to the list for retrieval.
				#return
			#else:
			    #if comparePackages(label, fakeLabel) <= 0:
				#return
		    #elif flags & rpm.RPMSENSE_EQUAL:

	depPkgList = []
 	if sense == rpm.RPMDEP_SENSE_REQUIRES:
 	    self.addReply("package %s needs '%s', version '%s'" % 
			  (name, needsName, needsVersion), 0, 1)
 	    dataList = []
	    if needsName in self.pkgList:
 	        tmp = urllib.quote_plus("solvedep %s" % needsName)
 	        if (needsVersion):
 		    tmp = tmp + urllib.quote_plus(" %s" % needsVersion)
            else:
                tmp = urllib.quote_plus("solvedep %s" % name)
                if (version):
                    tmp = tmp + urllib.quote_plus(" %s" % version)

 	    dataList.append("command=" + tmp)
	    dataList.append("username=" + urllib.quote_plus(self.username))
 	    dataList.append("auth=" + urllib.quote_plus(self.auth))
 	    dataList.append("state=" + urllib.quote_plus(self.state))
 	    dataStr = string.join(dataList, "&")
 	    #print "dataStr is", dataStr

	    (fp, boundary) = self.postAction(dataStr)
	    if fp == 0:
		return

	    if self.processState(fp, boundary):
		return

	    msg = multifile.MultiFile(fp, 0)

	    msg.push(boundary)
	    mmsg = mimetools.Message(msg)
	    try:
		for line in msg.readlines():
		    try:
			(dist, name, epoch, version, release) = string.split(line)
		    except:
			continue
		    tuple = "%s %s %s %s" % (name, epoch, version, release)
		    if not tuple in depPkgList:
			if not tuple in self.pkgList:
			    depPkgList.append(tuple)

		self.removeInstalledPackages(depPkgList, 1)
		self.removeSkipPackages(depPkgList)

		if self.gui:
		    for pkg in depPkgList:
			name = string.split(pkg)[0]
			self.gui.appendToPkgArea(name, NEW)

		self.retrieveHeaders(depPkgList)
		self.removeSkipFiles(depPkgList)
		self.pkgList = self.pkgList + depPkgList
			
	    except:
		# we maybe didn't get any information back, but
		# hopefully the problem will go away in one of
		# the next dependency passes.  One example case of
		# when this might happen is if a package has been added
		# that obsoletes a previously installed package, and
		# some library that package requires is getting upgraded,
		# and therefore it doesn't have the library it needs available
		# any more.  The next pass will remove that package because
		# of the new package that got added that obsoletes it.
		pass

	    fp.close()

 	if sense == rpm.RPMDEP_SENSE_CONFLICTS:
	    if not self.gui:
		self.addReply("%s, version %s conflicts with %s, version %s" %
			      (name, version, needsName, needsVersion), 0, 1)
	    else:
		self.gui.warning("%s, version %s conflicts with %s, version %s" %
				 (name, version, needsName, needsVersion))


    def retrievePackages(self):
	cid = self.setStatus(" Retrieving packages...")
	self.setProgress(0, 0)

	failedPkgList = []
	pkgCount = len(self.pkgList)
	for i in range(pkgCount):
	    if self.gui and self.gui.dieThreads:
		return

	    (name, epoch, version, release) = string.split(self.pkgList[i])
	    pkgName = "%s-%s-%s.%s.rpm" % (name, version, release, self.arch)
	    self.addReply(" Retrieving package %s..." % pkgName, 1, 1)

	    # may already be on disk
	    if os.access("%s/%s" % (self.cfg.readEntry("storageDir"), pkgName),
			    os.R_OK):
		self.addReply("already have it", 0, 1)
		if self.gui:
		    self.gui.updatePkgArea(name, DOWNLOAD)
		    self.setProgress(i, pkgCount)
		continue

	    # construct data to send
	    dataList = []
	    dataList.append("command=" + urllib.quote_plus("package %s" %
							   self.pkgList[i]))
	    dataList.append("username=" + urllib.quote_plus(self.username))
	    dataList.append("auth=" + urllib.quote_plus(self.auth))
	    dataList.append("state=" + urllib.quote_plus(self.state))
	    dataStr = string.join(dataList, "&")
	    #print "dataStr is", dataStr

	    (fp, boundary) = self.postAction(dataStr)
	    if fp == 0:
		return

	    if self.processState(fp, boundary):
		return
	    
	    msg = multifile.MultiFile(fp, 0)
	    msg.push(boundary)
	    mmsg = mimetools.Message(msg)
	    #print ("Content length of %s: %s" 
	    #   % (self.pkgList[i], mmsg.get("Content-Length")))
	    if mmsg.gettype() == "application/octet-stream":
		ofp = open("%s/%s" % 
			   (self.cfg.readEntry("storageDir"), pkgName), "w+")
		for line in mmsg.fp.readlines():
		    ofp.write(line)

		# the mime stuff sticks an extra newline on the
		# end of the file.  toss it.
		ofp.seek(-1, 1)
		ofp.truncate()
		ofp.close()
		self.addReply("retrieved.", 0, 1)
		if self.gui:
		    self.gui.updatePkgArea(name, DOWNLOAD)
	    else:
		self.addReply("FAILED.", 0, 1)
		failedPkgList.append(pkg)
		if self.gui:
		    self.gui.updatePkgArea(name, ERROR)

	    self.setProgress(i, pkgCount)
	    fp.close()

	for pkg in failedPkgList:
	    self.pkgList.remove(pkg)
	    name = string.split(pkg)[0]
	    if self.gui:
		self.gui.deletePkgArea(name)
	    
	self.unsetStatus(cid)
	
	
    def installPackages(self):
	cid = self.setStatus(" Installing packages...")
	self.setProgress(0, 0)

	try:
	    self.db = RPMDatabase(1, "/")
	except:
	    if not self.gui:
		self.addReply("Cannot open RPM database for writing, aborting.")
	    else:
		self.gui.error("Cannot open RPM database for writing, aborting.")

	    self.finishUp("Error reading RPM database")
	    return

	ts = rpm.TransactionSet("/", self.db.db)
	for pkg in self.pkgList:
	    if self.gui and self.gui.dieThreads:
		return

	    (name, epoch, version, release) = string.split(pkg)
	    pkgName = "%s-%s-%s.%s.rpm" % (name, version, release, self.arch)
	    
	    fd = os.open("%s/%s" % 
			 (self.cfg.readEntry("storageDir"),
			  pkgName), os.O_RDONLY)
	    (hdr, isSrc) = rpm.headerFromPackage(fd)
	    os.close(fd)
	    ts.add(hdr, "%s/%s" % (self.cfg.readEntry("storageDir"),
				   pkgName), "u")

	ts.order()

	if self.gui and self.gui.dieThreads:
	    return

	rc = ts.depcheck()
	if rc != None:
	    if not self.gui:
		self.addReply("Dependencies should be resolved, but they aren't!")
	    else:
		self.gui.error("Dependencies should be resolved, but they aren't!")

	    self.finishUp("Failed Red Hat Update Agent run")
	    return

	msg = ts.run(0, 0, self.callBack, "arg")
	if msg:
	    if not self.gui:
		self.addReply(msg)
	    else:
		self.gui.warning(msg)
		
	self.db.close()
	self.unsetStatus(cid)

    def deletePackageFiles(self):
	cid = self.setStatus(" Removing package files...")
	for pkg in self.pkgList:
	    (name, epoch, version, release) = string.split(pkg)
	    fileName = ("%s/%s-%s-%s.%s.rpm" % 
			(self.cfg.readEntry("storageDir"), name, 
			 version, release, self.arch))
	    os.remove(fileName)
	
	self.unsetStatus(cid)


    def startNetscape(self):
	os.system("rm -f %s/*.html" % self.cfg.readEntry("storageDir"))
		     
	cid = self.setStatus(" Building web page...")

	dataList = []
	commandStr = "command="
	commandStr = commandStr + urllib.quote_plus("display\n")
	for pkg in self.pkgList:
	    commandStr = commandStr + urllib.quote_plus("%s\n" % pkg)
	
	dataList.append(commandStr)
	dataList.append("username=" + urllib.quote_plus(self.username))
	dataList.append("auth=" + urllib.quote_plus(self.auth))
	dataList.append("state=" + urllib.quote_plus(self.state))
	dataStr = string.join(dataList,"&")

	try:
	    h = httplib.HTTP(self.cfg.readEntry("pkgServer"))
	except socket.error:
	    if self.gui:
		self.gui.error("Error connecting to package server.")
	    else:
		self.addReply('Error connecting to package server.')
	    self.finishUp("Failed Red Hat Update Agent run")
	    return
	    
	h.putrequest('POST', self.cfg.readEntry("pkgProg"))
	    
	h.putheader('Accept', '*/*')
	h.putheader('Content-length', "%s" % len(dataStr))
	h.putheader('Content-type', 'application/x-www-form-urlencoded')
	h.putheader('User-Agent', 'up2date/1.0')
	h.endheaders()
	
	h.send(dataStr)
	errcode, errmsg, headers = h.getreply()
	if errcode != 200:
	    if self.gui:
		self.gui.error("Error %s (%s) while receiving response from package server." % (errcode, errmsg))

	    else:
		self.addReply("Error %s (%s) while receiving response from package server." % (errcode, errmsg))
	    self.finishUp("Failed Red Hat Update Agent run")
	    return

	self.unsetStatus(cid)
	cid = self.setStatus(" Launching Netscape for package selection...")
	 
	import tempfile
	tempfile.tempdir = self.cfg.readEntry("storageDir")
	fn = tempfile.mktemp() + ".html"
	f = open(fn, "w+")
	f.write(h.getfile().read())
	f.close()

        npid = os.fork()
        if npid == 0:
            # child
            os.execvp("netscape", [ "netscape", "file://%s" % fn ])
	self.unsetStatus(cid)
        return (npid, fn)
    
