#
# This file is a portion of the Red Hat Update Agent
# Copyright (c) 1999 Red Hat, Inc.
#
# $Id: gui.py,v 1.31 1999/11/01 16:29:50 pbrown Exp $

from gtk import *
import urllib, sys, os, string
from threading import *
from client import Client
from gnome.ui import GnomeWarningDialog, GnomeErrorDialog, GnomeDialog, GnomeOkDialog
from gnome.uiconsts import *

from socket import *
from select import select

sock_path = "/var/run/up2date.sock"

# constants
NEW = 1
DEPEND = 2
INSTALLED = 3
ERROR = 4
DOWNLOAD = 5

class Gui:
    def __init__(self, auth):
	self.setupGuiElements()
	self.dieThreads = 0
	self.c = Client(auth, self)
	    

    def setupGuiElements(self):
	self.mainwin = GtkWindow(title="Red Hat Update Agent")

	vbox = GtkVBox()
	self.mainwin.add(vbox)

	logopix, logomask = create_pixmap_from_xpm(self.mainwin, 
						   None, "/usr/share/up2date/logo.xpm")
	try:
	    iconpix, iconmask = create_pixmap_from_xpm(self.mainwin,
						       None,
						       "/usr/share/pixmaps/redhat/shadowman-round-32.xpm")
	    self.mainwin.set_icon(iconpix, iconmask)
	except:
	    pass
	    
	vbox.pack_start(GtkPixmap(logopix, logomask), expand=FALSE)

	table = GtkTable()
	table.set_border_width(10)
	table.set_row_spacings(10)
	table.set_col_spacings(10)
	vbox.pack_start(table)

	self.pkgScroll = GtkScrolledWindow()
	self.pkgScroll.set_policy(hp=POLICY_AUTOMATIC, vp=POLICY_ALWAYS)
	table.attach(self.pkgScroll, 0, 1, 0, 4)

	self.pkgArea = GtkCList(cols=2, titles=("Package Name", "Status"))
	self.pkgArea.set_column_min_width(0, 150)
	self.pkgScroll.add(self.pkgArea)
	self.pkgScroll.set_usize(self.pkgArea.size_request()[0],
				 200)#)self.pkgArea.size_request()[1])

	# load up pixmaps for package area
	self.crosspix, self.crossmask = create_pixmap_from_xpm(self.mainwin,
							       None, "/usr/share/up2date/mini-cross.xpm")
	self.exclampix, self.exclammask = create_pixmap_from_xpm(self.mainwin,
								 None, "/usr/share/up2date/mini-exclam.xpm")
	self.defaultpix, self.defaultmask = create_pixmap_from_xpm(self.mainwin,
								   None, "/usr/share/up2date/mini-default.xpm")
	self.checkpix, self.checkmask = create_pixmap_from_xpm(self.mainwin,
							       None, "/usr/share/up2date/mini-check.xpm")
	self.lowerpix, self.lowermask = create_pixmap_from_xpm(self.mainwin,
							       None, "/usr/share/up2date/mini-lower.xpm")
	

	self.connectButton = GtkButton("Connect")
	self.connectButton.connect("clicked", self.connectAction)
	self.connectButton.set_usize(80,
				     self.connectButton.size_request()[1])
	table.attach(self.connectButton, 2, 3, 0, 1, yoptions=EXPAND)

	self.stopButton = GtkButton("Abort")
	self.stopButton.connect("clicked", self.reapThreads)
	self.stopButton.set_sensitive(FALSE)
	table.attach(self.stopButton, 2, 3, 1, 2, yoptions=EXPAND)

	self.configButton = GtkButton("Configure")
	self.configButton.connect("clicked", self.doConfig)
	table.attach(self.configButton, 2, 3, 2, 3, yoptions=EXPAND)

	quitButton = GtkButton("Quit")
	quitButton.connect("clicked", self.closedown)
	#quitButton.set_flags(CAN_DEFAULT)
	#quitButton.grab_default(TRUE)
	table.attach(quitButton, 2, 3, 3, 4, yoptions=EXPAND)

	hbox = GtkHBox()
	hbox.set_spacing(2)
	vbox.pack_start(hbox, expand=FALSE)

	self.progressBar = GtkProgressBar()
	self.progressBar.set_usize(100, -1)
	hbox.pack_start(self.progressBar)

	self.statusBar = GtkStatusbar()
	self.statusBar.set_spacing(5)
	self.statusBar.push(1, " Idle.")
	hbox.pack_start(self.statusBar)

	self.mainwin.connect("delete-event", mainquit)
	self.mainwin.connect("hide", self.closedown)
	self.mainwin.show_all()

    def closedown(self, *args):
        sock = socket(AF_UNIX, SOCK_STREAM)
        if sock > 0:
            try:
                sock.connect(sock_path)
                sock.close()
            except:
                # XXX what the hell should we do if we can not connect to the socket?
                pass
	gdk_flush()
	mainquit()

    def setCfgValues(self, *args):
	self.c.auth = self.c.cfg.readEntry("regKey")
	if self.c.auth == "":
	    self.c.auth = "null"
	self.c.username = self.c.cfg.readEntry("regName")
	
    def doConfig(self, *args):
	import configdlg
	cd = configdlg.ConfigDialog(self.c.cfg)
	cd.show_all()
	cd.connect("hide", self.setCfgValues)
    
    def reapThreads(self, *args):
	self.dieThreads = 1

    def cleanupGui(self):
	threads_enter()
	self.statusBar.push(1, " Idle.")
	self.progressBar.set_activity_mode(FALSE)
	self.progressBar.update(0)
	self.connectButton.set_sensitive(TRUE)
	self.configButton.set_sensitive(TRUE)
	self.stopButton.set_sensitive(FALSE)
	threads_leave()
	self.dieThreads = 0

    def refreshPkgArea(self):
	pass

    def appendToPkgArea(self, name, pixtype):
	threads_enter()
	if len(name) < 2:
	    print "WARNING: not adding a list to package area! (%s)" % name
	row = self.pkgArea.append((name,""))
	if pixtype == NEW:
	    self.pkgArea.set_pixmap(row, 1, 
				    self.defaultpix, self.defaultmask)
	elif pixtype == INSTALLED:
	    self.pkgArea.set_pixmap(row, 1,
				    self.checkpix, self.checkmask)
	elif pixtype == DEPEND:
	    self.pkgArea.set_pixmap(row, 1,
				    self.exclampix, self.exclammask)
	elif pixtype == DOWNLOAD:
	    self.pkgArea.set_pixmap(row, 1, 
				    self.lowerpix, self.lowermask)
	elif pixtype == ERROR:
	    self.pkgArea.set_pixmap(row, 1, 
				    self.crosspix, self.crossmask)
	threads_leave()

    def deletePkgArea(self, name):
	threads_enter()
	total = 0
	ok = 1
	while 1:
	    try:
		self.pkgArea.get_text(total, 0)
	    except:
		break
	    total = total + 1

	for i in range(total):
	    if name == self.pkgArea.get_text(i, 0):
		break

	# if we didn't find it we outta here
	if name != self.pkgArea.get_text(i, 0):
	    threads_leave()
	    return

	self.pkgArea.remove(i)
	threads_leave()


    def updatePkgArea(self, name, pixtype):
	threads_enter()
	total = 0
	ok = 1
	while 1:
	    try:
		self.pkgArea.get_text(total, 0)
	    except:
		break
	    total = total + 1

	for i in range(total):
	    if name == self.pkgArea.get_text(i, 0):
		break

	# if we didn't find it we outta here
	if name != self.pkgArea.get_text(i, 0):
	    threads_leave()
	    return

	
	if pixtype == NEW:
	    self.pkgArea.set_pixmap(i, 1, 
				    self.defaultpix, self.defaultmask)
	elif pixtype == INSTALLED:
	    self.pkgArea.set_pixmap(i, 1,
				    self.checkpix, self.checkmask)
	elif pixtype == DEPEND:
	    self.pkgArea.set_pixmap(i, 1,
				    self.exclampix, self.exclammask)
	elif pixtype == DOWNLOAD:
	    self.pkgArea.set_pixmap(i, 1, 
				    self.lowerpix, self.lowermask)
	elif pixtype == ERROR:
	    self.pkgArea.set_pixmap(i, 1, 
				    self.crosspix, self.crossmask)
	threads_leave()

    def connectAction(self, *args):
	t = Thread(target=self.mainThread)
	t.start()


    def warning(self, msg):
	threads_enter()
	mutex = Event()
	d = GnomeWarningDialog(msg, self.mainwin)
	d.connect("clicked", 
		  lambda dialog, button, mutex: mutex.set(), mutex)
	d.show_all()
	threads_leave()
	mutex.wait()

    def information(self, msg):
	threads_enter()
	mutex = Event()
	d = GnomeOkDialog(msg, self.mainwin)
	d.connect("clicked", 
		  lambda dialog, button, mutex: mutex.set(), mutex)
	d.show_all()
	threads_leave()
	mutex.wait()

    def error(self, msg):
	threads_enter()
	mutex = Event()
	d = GnomeErrorDialog(msg, self.mainwin)
	d.connect("clicked", 
		  lambda dialog, button, mutex: mutex.set(), mutex)
	d.show_all()
	threads_leave()
	mutex.wait()

    def cb (self, dialog, button, mutex):
	mutex.set()
	self.response = button
	dialog.destroy ()

    def delcb (self, widget, event, mutex):
	mutex.set ()
	self.response = 1
	widget.destroy ()

    def question(self, msg):
	threads_enter()
	mutex = Event()
	d = GnomeDialog("Update Agent Question", STOCK_BUTTON_OK, STOCK_BUTTON_CANCEL)
	vbox = d.vbox
	l = GtkLabel(msg)
	l.set_justify(JUSTIFY_LEFT)
	vbox.pack_start(l)
	d.connect("clicked", self.cb, mutex)
	d.connect("delete-event", self.delcb, mutex)
	d.set_parent(self.mainwin)
	d.show_all()
	threads_leave()
	mutex.wait()
	return self.response

    def yesno(self, msg):
	threads_enter()
	mutex = Event()
	d = GnomeDialog("Update Agent Question", STOCK_BUTTON_YES, STOCK_BUTTON_NO)
	vbox = d.vbox
	l = GtkLabel(msg)
	l.set_justify(JUSTIFY_LEFT)
	vbox.pack_start(l)
	d.connect("clicked", self.cb, mutex)
	d.connect("delete-event", self.delcb, mutex)
	d.set_parent(self.mainwin)
	d.show_all()
	threads_leave()
	mutex.wait()
	return self.response

    def mainThread(self):
	self.c.pkgList = []
 	threads_enter()
 	self.connectButton.set_sensitive(FALSE)
	self.configButton.set_sensitive(FALSE)
	self.pkgArea.clear()
	self.stopButton.set_sensitive(TRUE)
 	threads_leave()


	self.c.mailmsg = "Update Agent Run Summary:\n"
	self.c.mailmsg = self.c.mailmsg +  "------------------------\n\n"

 	self.c.getPackageList()

	if self.dieThreads:
	    self.cleanupGui()
	    return

 	self.c.removeInstalledPackages(self.c.pkgList,
 				       self.c.cfg.readEntry("retrieveNew"))

	if self.dieThreads:
	    self.cleanupGui()
	    return
	
 	skipped = self.c.removeSkipPackages(self.c.pkgList)
	if self.dieThreads:
	    self.cleanupGui()
	    return

	if len(skipped) > 0:
	    msg = "The following packages are available but were not selected\n"
	    msg = msg + "because they match filename patterns you have excluded:\n\n"
	    for s in skipped:
		name, epoch, version, release = string.split(s)
		msg = msg + "%s-%s-%s" % (name, version, release) + "\n"
	    #msg = msg + "\nDownload anyway?"

	    self.c.mailmsg = self.c.mailmsg + msg + "\n"

	    #if not self.yesno(msg):
		#self.c.roPkgList = skipped
	    self.warning(msg)

	if self.dieThreads:
	    self.cleanupGui()
	    return

        ns_pid = 0
        ns_file = ""

 	if len(self.c.pkgList) == 0:
	    self.warning("No packages available for retrieval\n"
			 " or installation.")
	    self.c.mailmsg = self.c.mailmsg + "No packages available for retrieval or installation.\n"
	    self.c.mailSummary()
	    self.cleanupGui()
	    return
	else:
            (ns_pid, ns_file) = self.c.startNetscape()

        if os.access(sock_path, 0):
            try:
                os.unlink(sock_path)
            except:
                if (os.access(sock_path, 0)):
                    # XXX: error box with can not set up socket
                    sys.exit(-1)

	cid = self.c.setStatus(" Processing response from Netscape...");

	if self.dieThreads:
	    self.cleanupGui()
	    return

        sock = socket(AF_UNIX, SOCK_STREAM)

	try:
	    sock.bind(sock_path)
	except:
	    self.error("Cannot create socket to communicate with Netscape.")
	    self.closedown()
            sys.exit(-1)

        sock.listen(1)
        conn, addr = sock.accept()       
        teststr = os.read(conn.fileno(), 1)
        if not teststr:
            # socket closed, that means that we might be quitting.
            try:
                conn.close()
                sock.close()
                os.unlink(sock_path)
            except:
                pass
            # get rid of the socket
            if os.access(sock_path, 0):
                os.unlink(sock_path)           
            # clean up netscape's file
            if ns_file and os.access(ns_file, 0):
                os.unlink(ns_file)
            # XXX: test that we are indeed quitting?
            sys.exit(0)

	if self.dieThreads:
	    self.cleanupGui()
	    return

        f = os.fdopen(conn.fileno())

        if (teststr == '\n'):
            stateData = ""
        else:
            stateData = teststr + f.readline()
        stateData = urllib.unquote_plus(stateData)
	self.c.state = stateData

	if self.dieThreads:
	    self.cleanupGui()
	    return

        self.c.pkgList = []
	for line in f.readlines():
            line = string.strip(line)
            if line == "":
                continue
	    name, epoch, version, release = string.split(line)
	    self.c.pkgList.append("%s %s %s %s" % (name, epoch, version, release))
	    self.appendToPkgArea(name, NEW)
        conn.close()
        sock.close()

	if self.dieThreads:
	    self.cleanupGui()
	    return
        
        # get rid of the socket
        if os.access(sock_path, 0):
            os.unlink(sock_path)
            
        # clean up netscape's file
        if ns_file and os.access(ns_file, 0):
            os.unlink(ns_file)

	if self.dieThreads:
	    self.cleanupGui()
	    return

	# unset "Processing response from Netscape" status
	self.c.unsetStatus(cid)

	# they may have been naughty and installed packages
	# between the two different stage runs.  Clean out any
	# that may be in the list but were installed in the interim.
 	self.c.removeInstalledPackages(self.c.pkgList,
 				       self.c.cfg.readEntry("retrieveNew"))

	if self.dieThreads:
	    self.cleanupGui()
	    return

	self.c.retrieveHeaders(self.c.pkgList)

	if self.dieThreads:
	    self.cleanupGui()
	    return

	skipped = self.c.removeSkipFiles(self.c.pkgList)

	if self.dieThreads:
	    self.cleanupGui()
	    return

	if len(skipped) > 0:
	    msg = "The following packages were not selected because they match\n"
	    msg = msg + "filename patterns you have excluded"
	    if self.c.cfg.readEntry("noReplaceConfig"):
		msg = msg + ", or contain locally\n"
		msg = msg + "modified configurations files"
	    msg = msg + ":\n\n"
	    for s in skipped:
		name, epoch, version, release = string.split(s)
		msg = msg + "%s-%s-%s" % (name, version, release) + "\n"
	    #msg = msg + "\nDownload anyway?"
	    self.c.mailmsg = self.c.mailmsg + msg

	    #if not self.yesno(msg):
		#self.c.roPkgList = skipped
	    self.warning(msg)


 	if len(self.c.pkgList) == 0:
	    self.warning("No packages available for retrieval\n"
			 " or installation.")
	    self.c.mailmsg = self.c.mailmsg + "No packages available for retrieval or installation.\n"
	    self.c.mailSummary()
	    self.cleanupGui()
	    return

 	self.c.dryRun()

	if self.dieThreads:
	    self.cleanupGui()
	    return

 	self.c.checkFreeDiskSpace()

	if self.dieThreads:
	    self.cleanupGui()
	    return

 	self.c.retrievePackages()

	if self.dieThreads:
	    self.cleanupGui()
	    return

 	if not self.c.cfg.readEntry("retrieveOnly"):
	    self.c.installPackages()

	if not self.c.cfg.readEntry("retrieveOnly"):
	    self.c.mailmsg = self.c.mailmsg + "The following packages were installed:\n\n"
	    msg = "All packages were installed successfully."
	else:
	    self.c.mailmsg = self.c.mailmsg + "The following packages were retrieved but not installed:\n\n"
	    msg = "All packages were successfully downloaded to %s.\nThey were not installed because of configuration settings." % self.c.cfg.readEntry("storageDir")

	for p in self.c.pkgList:
	    name, epoch, version, release = string.split(p)
	    self.c.mailmsg = self.c.mailmsg + "%s-%s-%s" % (name, version, release) + "\n"

	if self.c.cfg.readEntry("retrieveOnly"):
	    self.c.mailmsg = self.c.mailmsg + "\nThe packages have been stored in:\n\n %s\n\nfor manual installation.\n" % self.c.cfg.readEntry("storageDir")
	
	self.information(msg)
	self.c.mailSummary()

	if self.dieThreads:
	    self.cleanupGui()
	    return

 	if (not self.c.cfg.readEntry("retrieveOnly") and 
 	    not self.c.cfg.readEntry("keepAfterInstall")):
 	    self.c.deletePackageFiles()

	if self.dieThreads:
	    self.cleanupGui()
	    return

	self.cleanupGui()


if __name__ == "__main__":
    import config
    m = Gui("auth str")
    m.mainwin.connect("hide", mainquit)
    m.addText("this is some test text")
    mainloop()
